/****************************************************************** 
 * Copyright (c) 2003, Mario Hochreiter, Student of 
 * Software Engineering at FH-Hagenberg, Austria. 
 * Project member of the isac team at the Institute for 
 * Software-Technologie, Graz University of Technology, Austria. 
 * * 
 * Use is subject to license terms.
 ******************************************************************/
package isac.gui;

import isac.gui.treetable.CalcModelHierarchy;
import isac.gui.treetable.CalcModelNode;
import isac.gui.treetable.CalculationModel;
import isac.gui.treetable.CustomTreeCellEditor;
import isac.gui.treetable.CustomTreeCellRenderer;
import isac.interfaces.ICalcElement;
import isac.interfaces.IGuiToGui;
import isac.interfaces.IToGUI;
import isac.interfaces.IToUser;
import isac.useractions.EUIContext;
import isac.useractions.EUIElement;
import isac.useractions.IUIAction;
import isac.useractions.IUserAction;
import isac.useractions.UIAction;
import isac.useractions.UIActionOnCalcElement;
import isac.useractions.UserAction;
import isac.useractions.UserActionOnCalcElement;
import isac.useractions.UserActionOnPosition;
import isac.util.CalcChanged;
import isac.util.Formalization;
import isac.util.formulae.Assumptions;
import isac.util.formulae.CalcHead;
import isac.util.formulae.CalcFormula;
import isac.util.formulae.Position;
import isac.util.tactics.Tactic;
import isac.wsdialog.IWorksheetDialog;
import isac.wsdialog.DialogProtocolException;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.awt.image.BufferedImage;
import java.beans.PropertyVetoException;
import java.io.Serializable;
import java.rmi.RemoteException;
import java.util.HashMap;
import java.util.Locale;
import java.util.Vector;

import javax.swing.AbstractAction;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JDesktopPane;
import javax.swing.JInternalFrame;
import javax.swing.JLabel;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTree;
import javax.swing.KeyStroke;
import javax.swing.UIManager;
import javax.swing.event.CellEditorListener;
import javax.swing.event.ChangeEvent;
import javax.swing.event.InternalFrameEvent;
import javax.swing.event.InternalFrameListener;
import javax.swing.tree.DefaultTreeCellEditor;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.TreeModel;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;

import org.apache.log4j.Logger;

/**
 * CR070723 This should be WorksheetSwing, inheriting from Worksheet like a
 * WorksheetStub to be created; see java-tests/isac.gui.WorksheetSwing and
 * java-tests/isac.gui.WorksheetStub from a trial (where the Swing-code did 
 * not work properly).
 * 
 * <code>Worksheet</code> is the component that is responsible for the
 * interaction with the user while doing a calculation. Furthermore it acts as a
 * visual represantiation of mathematical formulas in a calculation tree, that
 * is beeing executed by the isac knowledge interpreter.
 * 
 * The <code>Worksheet</code> is the view of the <code>WorksheetDialog</code>.
 * 
 * The data, that the <code>WorksheetDialog</code> delivers, is representated
 * in the hierarchical structure <code>JTree</code> or on the
 * <code>CalcHeadPanel</code>.
 * 
 * @author Mario Hochreiter 2003
 * @version 0.1 15/09/2003
 * @author Alois Kirchsteiger 2004
 * @author Manuel Koschuch 2005
 */
public class Worksheet extends JPanel implements ActionListener,
        WindowListener, CellEditorListener, Serializable,
        InternalFrameListener, IGuiToGui, IGuiConstants {

    static final long serialVersionUID = 7877238175468368685L;

    private static final Logger logger = Logger.getLogger(Worksheet.class
            .getName());

    /***************************************************************************
     * Mouse Listener for handling mouse events in the worksheet tree
     * 
     */
    private MouseAdapter ml;

    /**
     * Context menu for each worksheet member
     */
    private JPopupMenu popupMenu;

    /***************************************************************************
     * The last point where the user clicked. Needed to show the Context Menu.
     */
    private Point mouse_point_;

    /***************************************************************************
     * Used for close listener and to add the Next and Auto buttons
     */
    private IToWorksheetUser windowApplication_;

    /***************************************************************************
     * The common actions that are applied to all nodes of the worksheet are
     * kept in this popu menu handler. The node specific commands are kept in
     * the handler of the node self.
     * 
     */
    private WorksheetPopupHandler commonActionsOnCalcElements_;

    // ****HashMaps for UIActions on CalcElments
    // HashMap containing the actions with the menu_items as key
    private HashMap<JComponent, IUIAction> dynamic_actions_by_menu_items_;

    // *************Contanier for dynamic buttons -- CONTEXT_CALCHEAD
    private Vector<Integer> dynamic_action_ids_with_context_calchead_;

    protected IToUser worksheet_rmi_1_;

    protected IToGUI worksheet_rmi_2_;

    // formula which will be edited if the dialog allows it
    protected CalcFormula edited_formula_;

    /** Model for the JTree */
    protected TreeModel model;

    /**
     * LK+WN0505 remove: only used for getting DialogIterators; WorksheetDialog
     * knows CalcTree, get Iterator from there
     * 
     * calc_tree_ removed
     */
    // protected CalcTree calc_tree_;
    protected CalcModelHierarchy calc_hierarchy_;

    /** Used to represent the model. */
    protected JTree tree_;

    /*
     * Renderer that is used to draw the tree cells
     */
    protected CustomTreeCellRenderer treeRenderer_;

    /**
     * The editor used to edit the cells in the tree
     */
    protected CustomTreeCellEditor treeEditor_;

    /** Dialog which is responsible for the communication with the kernel */
    protected IWorksheetDialog dialog_;

    /** Panel which is shows the calc_head_ if forced */
    protected CalcHeadPanel chp_;

    /**
     * NC060718: The following two fields : active_iterator_ and
     * selected_iterator_ should be deleted/replaced as soon as possible because
     * an CalcIterator represents a direct link to the mathematics engine.
     * Please do not use them in further extensions of the worksheet.
     * 
     * active_iterator_ removed selected_iterator_ removed
     */
    // protected ICalcIterator active_iterator_;
    // protected ICalcIterator selected_iterator_;
    private CalcModelNode edited_node_;

    private CalcModelNode selected_node_;

    protected CalcHead ch_;

    private JDesktopPane desktop_;

    private boolean calc_head_visible_;

    private JInternalFrame panel_frame_;

    /**
     * Create a <code>Worksheet</code> (and display the Specification if
     * requested by the viewStyle according to UserSettings). The given
     * <code>DGuide</code> is connected to the <code>Worksheet</code>.
     * 
     * @param dialogGuide
     *            The dialog guide that is connected to the Worksheet
     * @param desktop
     *            DesktopPane, the Worksheet is added to
     * @see ...{isac.wsdialog.WorksheetDialog#STARTFROM_*}
     */
    public Worksheet(IToWorksheetUser windowApplication,
            IWorksheetDialog dialogGuide, JDesktopPane desktop) {
        // LK 04.11.18 (1 line)
        desktop_ = desktop;
        calc_head_visible_ = false;
        windowApplication_ = windowApplication;

        // the following line registeres the worksheet as listener of the
        // window application in order to receive the window closing event
        windowApplication_.addCloseListener(this);

        // initialize the Containers for UIActions
        commonActionsOnCalcElements_ = new WorksheetPopupHandler();

        dynamic_actions_by_menu_items_ = new HashMap<JComponent, IUIAction>();
        dynamic_action_ids_with_context_calchead_ = new Vector<Integer>();

        setLayout(new BorderLayout());
        dialog_ = dialogGuide;
        try {
            worksheet_rmi_1_ = new WorksheetRMI(this);
            worksheet_rmi_2_ = new WorksheetRMI(this);

            dialog_.notifyGUIReady();

            dialog_.addDataChangeListener(worksheet_rmi_1_);
            dialog_.registerUIControlListener(worksheet_rmi_2_);
            // WN050421-AK: Worksheet should _ONLY_ know about DialogIterator
            // and use notifyUserAction for Calculation-related sutff.
            // But THIS system-generated call still requires calc_tree_ !!
            // WN050727 redesign startCalculation such that a CalcChanged
            // is thrown (thus same handling for root- and sub-pbl!)
            this.replaceCalcHeadPanelByWorksheet();
            dialog_.startCalculate();
        } catch (RemoteException e) {
            e.printStackTrace();
        }
    }

    protected void initMouseListener() {
        ml = new MouseAdapter() {

            public void mousePressed(MouseEvent e) {
                // ensure, that worksheet is always in the background, when
                // CalcHeadPanel is visible

                mouse_point_ = e.getPoint();
                TreePath selPath = tree_.getPathForLocation(e.getX(), e.getY());

                if (calc_head_visible_) {
                    moveToBack();
                } else {
                    if (selPath != null) {
                        selected_node_ = (CalcModelNode) selPath
                                .getLastPathComponent();

                        UserActionOnPosition changeContextFormula = new UserActionOnPosition(
                                EUIElement.UI_SOLVE_CHANGE_CONTEXT_FORMULA,
                                EUIContext.UI_CONTEXT_CALCELEMENTS,
                                selected_node_.getPosition());

                        try {
                            dialog_.notifyUserAction(changeContextFormula);
                        } catch (RemoteException e1) {
                            e1.printStackTrace();
                        } catch (DialogProtocolException e1) {
                            e1.printStackTrace();
                        }

                        if (e.getButton() == MouseEvent.BUTTON3
                                && (selected_node_.getPosition() != null)) {
                            showContextMenu(e);
                        }

                        if (e.getButton() == MouseEvent.BUTTON1
                                && e.getClickCount() == 2) {
                            editingStarted(e);
                        }
                    }

                }
            }
        };
        tree_.addMouseListener(ml);
    }

    /**
     * move Worksheet to back and CalcHeadPanel to front
     */
    private void moveToBack() {
        panel_frame_.toFront();
        try {
            panel_frame_.setSelected(true);
        } catch (PropertyVetoException e) {
            e.printStackTrace();
        }
        panel_frame_.validate();
        panel_frame_.updateUI();
    }

    /**
     * TreeView; attachs the Worksheet to the CalcHeadPanel
     */
    public void setTreeTableModel() {

        treeRenderer_ = new CustomTreeCellRenderer(this);
        treeEditor_ = new CustomTreeCellEditor(treeRenderer_, this);

        calc_hierarchy_ = new CalcModelHierarchy();
        model = new CalculationModel(calc_hierarchy_);
        tree_ = new JTree(model);
        tree_.setEditable(true);

        DefaultTreeCellRenderer dtcr = new DefaultTreeCellRenderer();

        Icon icon = UIManager.getIcon("OptionPane.errorIcon");

        System.out.println(icon.getClass());
        if (icon != null) {
            // Resize if necessary.
            JLabel tempLabel = new JLabel();
            BufferedImage img = new BufferedImage(icon.getIconWidth(), icon
                    .getIconHeight(), BufferedImage.TYPE_INT_RGB);

            Graphics g = img.getGraphics();
            g.setColor(Color.WHITE);
            g.fillRect(0, 0, icon.getIconWidth(), icon.getIconHeight());
            icon.paintIcon(tempLabel, g, 0, 0);

            ImageIcon resizedIcon = new ImageIcon(img.getScaledInstance(16, 16,
                    Image.SCALE_SMOOTH));

            dtcr.setLeafIcon(resizedIcon);
            dtcr.setOpenIcon(resizedIcon);
            dtcr.setClosedIcon(resizedIcon);
        }

        tree_.setCellRenderer(treeRenderer_);
        tree_
                .setCellEditor(new DefaultTreeCellEditor(tree_, dtcr,
                        treeEditor_));

        // / change it here
        treeEditor_.addCellEditorListener(this);

        tree_.getSelectionModel().setSelectionMode(
                TreeSelectionModel.SINGLE_TREE_SELECTION);
        tree_.setRootVisible(false);
        tree_.setShowsRootHandles(true);

        KeyStroke ks = KeyStroke.getKeyStroke(KeyEvent.VK_A,
                KeyEvent.CTRL_DOWN_MASK);
        KeyStroke edit_ks_ = KeyStroke.getKeyStroke(KeyEvent.VK_F2, 0);
        tree_.getInputMap().put(ks, "none");
        tree_.getInputMap().put(edit_ks_, "startEditing");

        tree_.getActionMap().put("startEditing",
                new AbstractAction("startEditing") {

                    static final long serialVersionUID = 2301979180001L;

                    public void actionPerformed(ActionEvent evt) {
                        TreePath sel_path = tree_.getSelectionPath();

                        if (sel_path == null)
                            return;

                        edited_node_ = (CalcModelNode) sel_path
                                .getLastPathComponent();
                        Rectangle ed_node_rect = tree_.getPathBounds(sel_path);

                        MouseEvent me = new MouseEvent(tree_, 0, 0, 0,
                                ed_node_rect.x + 1, ed_node_rect.y + 1, 2,
                                false);
                        editingStarted(me);
                    }
                });

        add(new JScrollPane(tree_), BorderLayout.CENTER);

        tree_.updateUI();
        tree_.repaint();

        this.initMouseListener();
    }

    /**
     * replace the CalcHeadPanel by the Worksheet FIXXXME.WN050727 streichen, +
     * das zugeh"orige setzen streichen!
     */
    private void replaceCalcHeadPanelByWorksheet() {
        this.setTreeTableModel();
        this.validate();// WN050421???
    }

    /**
     * 
     * 
     * Description of the method in HTML format, if a link can be given in the
     * running text do this with {@link
     * fully.qualified.Classname#methodName(fully.qualified.Paramclass)}
     * 
     * @param <paramname>
     *            <paramdescription>
     * @return <return value>
     * @exception <exception>
     *                <description when it is thrown>
     * @see <fully.qualified.Classname#methodName(param-classes)>
     */
    public void startSolving() {
        try {
            // dialog.newCalculation(ch);
            System.out.println("in startSolving()");
            chp_.fillCalcHeadFromTextFields(ch_);
            if (logger.isInfoEnabled())
                logger.info("WS->DG: UI_SPECIFY_TO_SOLVE");
            if (dialog_.notifyUserAction(new UserAction(
                    EUIElement.UI_SPECIFY_TO_SOLVE))) {
                replaceCalcHeadPanelByWorksheet();
            } else {
                JOptionPane.showMessageDialog(this,
                        "Please enter a correct CalcHead",
                        "CalcHead not correct", JOptionPane.WARNING_MESSAGE);
            }
        } catch (RemoteException e1) {
            logger.error("actionPerformed", e1.detail);
        } catch (DialogProtocolException e2) {
            logger.error("actionPerformed", e2);
            logger.error("e2" + e2);
        }
    }

    public void compCalcHead() {
        try {
            chp_.fillCalcHeadFromTextFields(ch_);
            dialog_.notifyUserAction(new UserAction(
                    EUIElement.UI_SPECIFY_COMPLETE_CALCHEAD));
            chp_.fillTextFieldsFromCalcHead(ch_);
        } catch (Exception e1) {
            logger.error(e1);
        }
    }

    public void checkCalcHead() {
        try {
            if (logger.isDebugEnabled())
                logger.debug(dialog_);
            dialog_.notifyUserAction(new UserAction(
                    EUIElement.UI_SPECIFY_CHECK_CALCHEAD));
            chp_.fillTextFieldsFromCalcHead(ch_);
        } catch (Exception e1) {
            logger.error(e1);
        }
    }

    /**
     * Adds a Button to the CalcHeadPanel
     * 
     * @param action
     *            the action associated with the button
     */
    private void addCalcHeadButton(IUIAction action) {
        JButton dynamic_button = new JButton(action.getLabel());
        dynamic_button.addActionListener(this);
        dynamic_actions_by_menu_items_.put(dynamic_button, action);

        chp_.getButtonPanel().add(dynamic_button);
        super.validate();
    }

    /**
     * Shows CalcHeadPanel. Creates a new InternalFrame on the desktopPane the
     * worksheet is added to and ensures, that this frames stays in front of the
     * Worksheet as long as it is opened
     * 
     * @param calc_head
     *            CalcHead to display in Panel
     * 
     */
    private void openCalcHeadPanel(CalcHead calc_head) {

        panel_frame_ = new JInternalFrame("CalcHeadPanel", true, // resizable
                false, // closable
                false, // maximisable
                false); // minimisable

        panel_frame_.setResizable(true);
        panel_frame_.setClosable(false);
        panel_frame_.setMaximizable(false);

        chp_ = new CalcHeadPanel(calc_head);

        panel_frame_.getContentPane().add(chp_);

        panel_frame_.setPreferredSize(new Dimension(CHP_INITIAL_SIZE_X,
                CHP_INITIAL_SIZE_Y));
        panel_frame_.addInternalFrameListener(this);

        panel_frame_.pack();

        panel_frame_.validate();

        // NC 060616
        if (mouse_point_ != null)
            panel_frame_.setLocation(mouse_point_.x + CHP_X_POSITION_OFFSET,
                    mouse_point_.y + CHP_Y_POSITION_OFFSET);
        else
            panel_frame_.setLocation(CHP_X_POSITION_OFFSET,
                    CHP_Y_POSITION_OFFSET);

        panel_frame_.setVisible(true);

        chp_.getModelPanel().addFocusLostHandler(this);
        chp_.getSpecificationPanel().addRadioButtonHandler(this);

        desktop_.add(panel_frame_);
        calc_head_visible_ = true;

        this.validate();

        moveToBack();

        tree_.setEditable(!calc_head_visible_);
    }

    public void registerActionWithListener(JMenuItem item, IUIAction action) {
        dynamic_actions_by_menu_items_.put(item, action);
    }

    // ------------- Implementation of ActionListener ----------------
    public void actionPerformed(ActionEvent e) {
        UIAction action = (UIAction) dynamic_actions_by_menu_items_.get(e
                .getSource());

        EUIContext context = action.getContext();

        try {
            switch (context) {
            // if CalcHeadPanel Button is pressed: alwas send calchead created
            // by reading all Fields in the CalcHeadPanel
            case UI_CONTEXT_CALCHEAD:
                CalcHead calc_head = new CalcHead();
                chp_.fillCalcHeadFromTextFields(calc_head);
                dialog_.notifyUserAction(new UserActionOnCalcElement(action
                        .getAction(), calc_head));
                break;
            case UI_CONTEXT_ONEELEMENT:
                dialog_.notifyUserAction(new UserActionOnCalcElement(action
                        .getAction(), selected_node_.getCalcElement()));
                break;

            case UI_CONTEXT_APPLICABLE_TACTIC:
                dialog_.notifyUserAction(new UserActionOnCalcElement(action
                        .getAction(), action.getContext(),
                        action.getPosition(), ((UIActionOnCalcElement) action)
                                .getCalcElement()));
                break;
            default:
                dialog_.notifyUserAction(new UserActionOnPosition(action
                        .getAction(), EUIContext.UI_CONTEXT_NO_CONTEXT,
                        selected_node_.getPosition()));
                break;
            }

        } catch (RemoteException ex1) {
            ex1.printStackTrace();
        } catch (DialogProtocolException ex2) {
            ex2.printStackTrace();
        }
    }

    public void calcChanged(CalcChanged event) {
        ((CalculationModel) model).calcChanged(event);
        for (int i = 0; i < tree_.getRowCount(); i++) {// WN050428 ???+DG
            // filters
            tree_.expandRow(i);
        }
        tree_.setEditable(true);
        try {
            dialog_.notifyGUIReady();
        } catch (RemoteException e) {
            e.printStackTrace();
        }
    }

    private void showContextMenu(MouseEvent click) {
        TreePath selPath = tree_.getPathForLocation(click.getX(), click.getY());
        tree_.setSelectionPath(selPath);

        edited_node_ = (CalcModelNode) selPath.getLastPathComponent();

        if (isClickInTactic(click)) {
            // / NC060705: the Applicable Tactics menu is not sent with
            // CalcChanged event, but is
            // / loaded on demand each time a user right clicks in the tactic
            // field
            //
            // / popupMenu = cnode.getTacticContextMenu( this );
            // / send a UserAction UI_SOLVE_GET_APPLICABLE_TACTICS to Worksheet

            try {
                dialog_.notifyUserAction(new UserActionOnPosition(
                        EUIElement.UI_SOLVE_GET_APPLICABLE_TACTICS,
                        EUIContext.UI_CONTEXT_ONEELEMENT, selected_node_
                                .getPosition()));
            } catch (RemoteException e) {
                e.printStackTrace();
            } catch (DialogProtocolException e) {
                e.printStackTrace();
            }

            return;
        } else {
            popupMenu = commonActionsOnCalcElements_.buildContextMenu(this);
            popupMenu = edited_node_.getFormulaContextMenu(this, popupMenu);
        }

        if (popupMenu != null)
            popupMenu.show(click.getComponent(), click.getX(), click.getY());
        else
            System.out.println("Error: I have no popup menu to show !");
    }

    /**
     * Every tree cell has the formula and tactic part. This method checks
     * whether the given click is in the tactic part or the formula part of the
     * cell.
     * 
     * @param click -
     *            The click to check
     * @return true - when the tactic part of the cell was clicked
     */
    public boolean isClickInTactic(MouseEvent click) {

        TreePath selPath = tree_.getPathForLocation(click.getX(), click.getY());
        CalcModelNode current_node = (CalcModelNode) selPath
                .getLastPathComponent();

        if (current_node == null)
            return false;

        int current_row = tree_.getRowForPath(selPath);
        Rectangle target_rect = tree_.getPathBounds(selPath);

        CustomTreeCellRenderer ctcr = (CustomTreeCellRenderer) tree_
                .getCellRenderer();
        CustomTreeCellRenderer rendered_cell = (CustomTreeCellRenderer) ctcr
                .getTreeCellRendererComponent(tree_, current_node, true, true,
                        true, current_row, true);

        int in_component_y = (mouse_point_.y - target_rect.y);
        Dimension formula_rect = rendered_cell.getFormulaField()
                .getPreferredSize();

        if (in_component_y <= formula_rect.height)
            return false;
        else
            return true;
    }

    /**
     * Called when the user starts to edit the active cell. The cell contains
     * formula and tactic. Depending on where the user clicked, the apropriate
     * subcell is edited.
     * 
     */
    public void editingStarted(MouseEvent e) {

        if (isClickInTactic(e))
            treeEditor_.activateSubCell(CustomTreeCellEditor.SUBCELL_TACTIC);
        else
            treeEditor_.activateSubCell(CustomTreeCellEditor.SUBCELL_FORMULA);

        edited_node_ = (CalcModelNode) tree_.getSelectionPath()
                .getLastPathComponent();
        tree_.startEditingAtPath(tree_.getSelectionPath());
    }

    /**
     * if the user has canceled the editing, the worksheet should inform the
     * dialogGuide
     * 
     * @see javax.swing.event.CellEditorListener#editingCanceled(javax.swing.event.ChangeEvent)
     */
    public void editingCanceled(ChangeEvent e) {

    }

    /**
     * if the user has changed the value of a node, it should be altered
     * 
     * @see javax.swing.event.CellEditorListener#editingStopped(javax.swing.event.ChangeEvent)
     */
    public void editingStopped(ChangeEvent e) {
        String o = (String) tree_.getCellEditor().getCellEditorValue();
        try {
            CalcFormula formula = null;

            if (edited_node_ != null)
                formula = edited_node_.getFormula();

            if (formula == null) { // append formula
                // appends after the currently active formula
                System.out.println("Add new formula ");
                edited_formula_ = new CalcFormula();
                edited_formula_.setText(o);
                if (logger.isInfoEnabled())
                    logger.info(" WS->DG: UI_SOLVE_APPEND_USER_FORMULA");
                dialog_.notifyUserAction(new UserAction(
                        EUIElement.UI_SOLVE_APPEND_USER_FORMULA));
            } else { // replace formula
                System.out.println("Replace formula " + formula.getPosition());
                if (logger.isInfoEnabled())
                    logger.info(" WS->DG: UI_SOLVE_MOVE_ACTIVE_FORMULA");
                dialog_
                        .notifyUserAction(new UserActionOnPosition(
                                EUIElement.UI_SOLVE_MOVE_ACTIVE_FORMULA,
                                EUIContext.UI_CONTEXT_ONEELEMENT, formula
                                        .getPosition()));
                edited_formula_ = new CalcFormula();
                edited_formula_.setText(o);
                edited_formula_.setPosition(formula.getPosition());
                if (logger.isInfoEnabled())
                    logger.info(" WS->DG: UI_SOLVE_EDIT_ACTIVE_FORMULA");
                dialog_.notifyUserAction(new UserAction(
                        EUIElement.UI_SOLVE_EDIT_ACTIVE_FORMULA));
            }
        } catch (Exception e1) {
            e1.printStackTrace();
        }
        System.out.println(o);
    }

    /**
     * @param action
     * @deprecated in favour of doUIAction(IUIAction action) with introduction
     *             of IToGUI
     */
    public boolean doUIAction(IUserAction action) {
        return false;
    }

    /**
     * add an interactive element on the gui depending on the context
     * 
     * @see isac.interfaces.IToGUI#addUIElement(isac.wsdialog.IUIAction)
     */
    public void addUIElement(IUIAction action) {
        System.out.println("WSD->WS addUIElement : action = "
                + action.getAction() + " Context = " + action.getContext());

        EUIContext context = action.getContext();
        switch (context) {
        case UI_CONTEXT_CALCULATION:
            windowApplication_.addToMenuBar(action);
            break;
        case UI_CONTEXT_CALCELEMENTS:
            commonActionsOnCalcElements_.addAction(action);
            break;

        case UI_CONTEXT_ONEELEMENT:
            Position action_position = ((UIAction) action).getPosition();
            CalcModelNode action_node = calc_hierarchy_.getNodeForValue(
                    action_position, calc_hierarchy_.getRoot());

            if (action_node != null)
                action_node.addDynamicAction(action);
            break;

        case UI_CONTEXT_CALCHEAD:
            if (!dynamic_action_ids_with_context_calchead_
                    .contains(new Integer(action.getAction().ordinal()))) {
                addCalcHeadButton(action);
            }
            break;

        case UI_CONTEXT_APPLICABLE_TACTIC:
            Position tactic_position = ((UIActionOnCalcElement) action)
                    .getPosition();
            CalcModelNode tactic_node = calc_hierarchy_.getNodeForValue(
                    tactic_position, calc_hierarchy_.getRoot());

            if (tactic_node != null)
                tactic_node.addApplicableTactic(action);
            break;
        }
    }

    /**
     * remove an interactive element from the gui
     * 
     * @see isac.interfaces.IToGUI#removeUIElement(isac.wsdialog.IUIAction)
     */
    public boolean removeUIElement(IUIAction action) {
        if (logger.isInfoEnabled())
            logger.info("WD->WS removeUIElement : action = "
                    + action.getAction() + " Context = " + action.getContext());

        EUIContext context = action.getContext();
        switch (context) {
        // TOODO: build all UIElements from wsdialog with this method

        case UI_CONTEXT_CALCULATION:
            return windowApplication_.removeFromMenuBar(action);
        case UI_CONTEXT_CALCELEMENTS:
            break;
        case UI_CONTEXT_ONEELEMENT:
            break;
        case UI_CONTEXT_CALCHEAD:
            return windowApplication_.removeFromMenuBar(action);
        }
        return false;
    }

    /**
     * notify UserActions which come from outside the Worksheet, i.e.
     * WindowApplication#actionPerformed
     */
    public void notifyUserAction(IUserAction action) {
        try {
            switch (action.getContext()) {
            case UI_CONTEXT_CALCHEAD:
                CalcHead calc_head = new CalcHead();
                chp_.fillCalcHeadFromTextFields(calc_head);
                dialog_.notifyUserAction(new UserActionOnCalcElement(action
                        .getActionID(), calc_head));
                break;
            default:
                dialog_.notifyUserAction(action);
                break;

            // //all other contexts are irrelevant here
            // case EUIContext.UI_CONTEXT_CALCELEMENTS:
            // case EUIContext.UI_CONTEXT_ONEELEMENT:
            // case EUIContext.UI_CONTEXT_SWITCH_PHASES:
            }
        } catch (RemoteException e) {
            e.printStackTrace();
        } catch (DialogProtocolException e) {
            e.printStackTrace();
        }
    }

    /**
     * Services the UI_SOLVE_TACTIC_APPLIED UIAction. This UIAction is thrown
     * from WorksheetDialog as a direct response to the IUAction of the SAME
     * type. This method is called only if the returned element is indeed Tactic
     * and if the element is not null.
     * 
     * @param ta -
     *            the tactic applied
     * 
     */
    private void doTacticApplied(Tactic ta) {
        TreePath sel_path = tree_.getSelectionPath();

        if (sel_path == null)
            return;

        edited_node_ = (CalcModelNode) sel_path.getLastPathComponent();
        edited_node_.setTacticString(ta.toSMLString());

        model.valueForPathChanged(sel_path, edited_node_);
    }

    /**
     * Services the UI_SOLVE_SHOW_ASSUMPTIONS UIAction. This UIAction is thrown
     * from WorksheetDialog. This method is called only if the returned element
     * is indeed Assumptions and if the element is not null. The data is written
     * in the tacticString field of the edited node.
     * 
     * 
     * @param ass -
     *            the new Assumption
     * 
     */
    private void doShowAssumptions(Assumptions asm) {
        TreePath sel_path = tree_.getSelectionPath();

        if (sel_path == null)
            return;

        edited_node_ = (CalcModelNode) sel_path.getLastPathComponent();
        edited_node_.setAssumptionString(asm.toSMLString());

        model.valueForPathChanged(sel_path, edited_node_);
    }

    /**
     * Services the UI_SOLVE_SHOW_APPLICABLE_TACTICS UIAction. This UIAction is
     * thrown from WorksheetDialog.
     * 
     * When the user clicks on the tactic field with the right mouse click, the
     * UI_SOLVE_GET_APPLICABLE_TACTICS is sent to the WSDialog, the WSDialog
     * retrieves the tactics from the math engine and responds with
     * UI_SOLVE_SHOW_APPLICABLE_TACTICS to the Worksheet
     * 
     */
    private void doShowApplicableTactics() {
        popupMenu = edited_node_.getTacticContextMenu(this);

        popupMenu.show(this, mouse_point_.x, mouse_point_.y);
    }

    /**
     * perform the UIAction requested from the WorksheetDialog
     * 
     * @see isac.interfaces.IToGUI#doUIAction(isac.wsdialog.IUIAction)
     */
    public void doUIAction(IUIAction action) {
        if (logger.isInfoEnabled())
            logger.info("WD->WS doUIAction : action = " + action.getAction()
                    + " Context = " + action.getContext());
        switch (action.getContext()) {
        // TOODO: complete groups of UIElements
        case UI_CONTEXT_CALCULATION:
            break;
        case UI_CONTEXT_CALCELEMENTS:
            switch (action.getAction()) {
            case UI_SOLVE_TACTIC_APPLIED:
                if (action instanceof UIActionOnCalcElement) {
                    ICalcElement ice = ((UIActionOnCalcElement) action)
                            .getCalcElement();

                    if (ice == null) {
                        System.out.println("Tactic is null");
                        return;
                    }

                    if (ice instanceof Tactic)
                        doTacticApplied((Tactic) ice);
                    else
                        System.out.println("Not my tactic : "
                                + ice.toSMLString() + " is "
                                + ice.getClass().getName());
                } else {
                    System.out.println("Not my action : "
                            + action.getClass().getName());
                }

                break;
            case UI_SOLVE_SHOW_ASSUMPTIONS:
                if (action instanceof UIActionOnCalcElement) {
                    ICalcElement ice = ((UIActionOnCalcElement) action)
                            .getCalcElement();

                    if (ice == null) {
                        System.out.println("Assumption is null");
                        return;
                    }

                    if (ice instanceof Assumptions)
                        doShowAssumptions((Assumptions) ice);
                    else
                        System.out.println("Not my assumption : "
                                + ice.toSMLString() + " is "
                                + ice.getClass().getName());
                }
                break;
            }
            break;
        case UI_CONTEXT_ONEELEMENT:
            switch (action.getAction()) {
            case UI_SOLVE_SHOW_APPLICABLE_TACTICS:
                doShowApplicableTactics();
                break;
            case UI_DO_EDIT_FORMULA:
                try {
                    dialog_.notifyUserAction(new UserActionOnCalcElement(
                            EUIElement.UI_SOLVE_EDIT_ACTIVE_FORMULA_COMPLETE,
                            edited_formula_));
                } catch (RemoteException e) {
                    e.printStackTrace();
                } catch (DialogProtocolException e) {
                    e.printStackTrace();
                }
                break;
            case UI_DO_APPEND_FORMULA:
                try {
                    dialog_.notifyUserAction(new UserActionOnCalcElement(
                            EUIElement.UI_SOLVE_EDIT_ACTIVE_FORMULA_COMPLETE,
                            edited_formula_));
                } catch (RemoteException e) {
                    e.printStackTrace();
                } catch (DialogProtocolException e) {
                    e.printStackTrace();
                }
                break;
            case UI_SOLVE_CHANGE_CONTEXT_FORMULA:
                UIAction ccf_action = (UIAction) action;
                selected_node_ = calc_hierarchy_.getNodeForValue(ccf_action
                        .getPosition(), calc_hierarchy_.getRoot());
                if (!calc_head_visible_) {
                    TreePath sel_path = calc_hierarchy_
                            .getPathToNode(selected_node_);
                    tree_.setSelectionPath(sel_path);
                    tree_.updateUI();
                }
                break;
            default:
                break;
            }
            break;
        case UI_CONTEXT_SWITCH_PHASES:
            if (!calc_head_visible_) {
                CalcHead calc_head = (CalcHead) (((UIActionOnCalcElement) action)
                        .getCalcElement());
                openCalcHeadPanel(calc_head);
            } else {
                closeCalcHeadPanel();
            }
            break;
        case UI_CONTEXT_CALCHEAD:
            chp_
                    .fillTextFieldsFromCalcHead((CalcHead) (((UIActionOnCalcElement) action)
                            .getCalcElement()));
            break;
        }
    }

    private void closeCalcHeadPanel() {
        panel_frame_.dispose();
        tree_.setEditable(true);
    }

    public void internalFrameOpened(InternalFrameEvent e) {
    }

    public void internalFrameClosing(InternalFrameEvent e) {
    }

    public void internalFrameClosed(InternalFrameEvent e) {

        if (e.getSource() == panel_frame_) {
            calc_head_visible_ = false;
            tree_.setEditable(!calc_head_visible_);
            return;
        }

        try {
            dialog_.notifyUserAction(new UserAction(
                    EUIElement.UI_CLOSE_WORKSHEET));
        } catch (RemoteException e1) {
            e1.printStackTrace();
        } catch (DialogProtocolException e1) {
            e1.printStackTrace();
        }

        windowApplication_.removeCloseListener(this);
    }

    public void internalFrameIconified(InternalFrameEvent e) {
    }

    public void internalFrameDeiconified(InternalFrameEvent e) {
    }

    public void internalFrameActivated(InternalFrameEvent e) {

        try {
            dialog_.notifyUserAction(new UserAction(
                    EUIElement.UI_ACTIVE_WORKSHEET));
        } catch (RemoteException e1) {
            e1.printStackTrace();
        } catch (DialogProtocolException e1) {
            e1.printStackTrace();
        }
    }

    public void internalFrameDeactivated(InternalFrameEvent e) {

        if (e.getSource() == panel_frame_) {
            System.out.println("CalcHead deactivated");
            moveToBack();
            return;
        }

    }

    public void windowActivated(WindowEvent arg0) {
        return;
    }

    public void windowClosed(WindowEvent arg0) {
        windowApplication_.removeCloseListener(this);
    }

    // / the method gets the notifications from WindowApplication main Window
    public void windowClosing(WindowEvent arg0) {
        calc_head_visible_ = false;
        try {
            dialog_.notifyUserAction(new UserAction(
                    EUIElement.UI_CLOSE_WORKSHEET));
        } catch (RemoteException e1) {
            e1.printStackTrace();
        } catch (DialogProtocolException e1) {
            e1.printStackTrace();
        }

    }

    public void windowDeactivated(WindowEvent arg0) {

    }

    public void windowDeiconified(WindowEvent arg0) {

    }

    public void windowIconified(WindowEvent arg0) {

    }

    public void windowOpened(WindowEvent arg0) {

    }
}