package isac.users;

import java.net.URL;
import java.rmi.RemoteException;
import java.sql.Connection;
import java.sql.Date;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.HashMap;

import isac.browserdialog.ExampleDialog;
import isac.interfaces.ICalcIterator;
import isac.useractions.EUIElement;
import isac.useractions.IUIAction;
import isac.useractions.IUserAction;
import isac.useractions.UserActionOnCalcElement;
import isac.useractions.UserActionOnLink;
import isac.util.CalcChanged;
import isac.util.CalcMessage;
import isac.util.ObjectManagerPaths;
import isac.util.formulae.CalcHead;
import isac.util.formulae.Formula;

/**
 * Records steps of the user. There is one instance per session. WN070908 the
 * current version of this class assumes that the front-end is frozen until any
 * request is answered.
 * 
 * @see UserLogger#current_step_
 * 
 * A step is an action of the user which can be assigned a unique
 * <code>SuccessInfo</code>. Frequently this is a single
 * <code>UserAction</code>; but there are compound-steps like doing a whole
 * example (finishing it successfully or not), or repeating input a formula
 * (ending with a successful input or not), etc.
 * 
 * <code>UserActions</code> on the KnowledgeBrowsers (i.e.<code>BrowserFrames</code>)
 * do not have a <code>SuccessInfo</code>, if there is no context to a
 * <code>Worksheet</code>.
 * 
 * @see isac.users.UserRecord for the fields of a step's record.
 * 
 * @author Christian Ropposch
 */
public class UserLogger {
    private boolean log_to_database_enabled_ = false;

    private Connection connection_ = null;

    private String session_id_;

    private User user_;// FIXME.WN071011 user_name_ would be sufficient

    /**
     * @deprecated quick and dirty made Compod run ...
     */
    private ExampleDialog exp_dialog_;

    /**
     * Holds the <code>UserRecord</code> partially filled by
     * UserLogger#logUserAction waiting to be completed by
     * UserLogger#logCalcChanged.
     */
    private UserRecord current_step_ = new UserRecord();

    // private HashMap<String, UserRecord> worksheet_map_user_;

    /** There is one SuccessInfo per Worksheet. */
    // private HashMap<String, SuccessInfo> worksheet_map_success_;
    public UserLogger(String session_id, User user) {
        session_id_ = session_id;
        user_ = user;
        // worksheet_map_user_ = new HashMap<String, UserRecord>();
        // worksheet_map_success_ = new HashMap<String, SuccessInfo>();

        log_to_database_enabled_ = ObjectManagerPaths.LOGGER_DATABASE_ENABLED;

        if (log_to_database_enabled_) {
            try {
                connectDatabase(); // establish database-connection
                current_step_.setAction(user.getUsername(), session_id,
                        null /* no dialog */,
                        isac.useractions.EUIElement.LO_START_SESSION,
                        "TODO user-url", null);
                // TODO user-url from isacweb, shall come with new login
                currentStepToDatabase();
            } catch (SQLException e) {
                e.printStackTrace();
            }
        }
    }

    protected void finalize() {
        if (log_to_database_enabled_) {
            try {
                if (!(connection_.isClosed()))
                    connection_.close();
            } catch (SQLException e) {
                e.printStackTrace();
            }
        }
    }

    private void connectDatabase() throws SQLException {
        try {
            Class.forName("com.mysql.jdbc.Driver");
        } catch (ClassNotFoundException e) {
            e.printStackTrace();
        }

        connection_ = DriverManager.getConnection("jdbc:mysql://"
                + ObjectManagerPaths.LOGGER_DATABASE_PATH + ":"
                + ObjectManagerPaths.LOGGER_DATABASE_PORT + "/"
                + ObjectManagerPaths.LOGGER_DATABASE_NAME
                + "?useUnicode=true&characterEncoding=UTF-8",
                ObjectManagerPaths.LOGGER_DATABASE_USER,
                ObjectManagerPaths.LOGGER_DATABASE_PASSWORD);
    }

    private void writeToDatabase(String user_name, String session, String step,
            String worksheet, String position, String formula_from,
            String formula_to, String argument, String substitution,
            String dialog, Float success) {
        if (log_to_database_enabled_) {
            try {
                if (connection_.isClosed())
                    this.connectDatabase();

                String sql_stmt = "insert into UserLogger(UserName, Session, Dialog, Step, Worksheet, Position, FormulaFrom, FormulaTo, "
                        + "Argument, Substitution, Success) "
                        + "values (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)";
                PreparedStatement statement = connection_.prepareStatement(
                        sql_stmt, PreparedStatement.RETURN_GENERATED_KEYS);
                statement.setString(1, user_name);
                statement.setString(2, session);
                statement.setString(3, dialog); // dialog
                statement.setString(4, step); // step
                statement.setString(5, worksheet); // worksheet
                statement.setString(6, position); // position
                statement.setString(7, formula_from); // formulafrom
                statement.setString(8, formula_to); // formulato
                statement.setString(9, argument); // argument
                statement.setString(10, substitution); // substitution
                if (success == null)
                    success = (float) 0.0;
                statement.setFloat(11, success); // success
                statement.execute();
            } catch (SQLException e) {
                e.printStackTrace();
            }
        }
    }

    private void currentStepToDatabase() {
        writeToDatabase(current_step_.getUser_name(), current_step_
                .getSession(), current_step_.getStep(), current_step_
                .getWorksheet(), current_step_.getPosition(), current_step_
                .getFormula_from(), current_step_.getFormula_to(),
                current_step_.getArgument(), current_step_.getSubst(),
                current_step_.getDialog(), current_step_.getSucces());
        current_step_.reset();
    }

    /**
     * Log a UserAction which are <em>not</em> related to a
     * <code>CalcChanged</code> and <em>not</em> related to a
     * <code>ContexProvider</code>.
     * 
     * If related to one of these then the data are stored for retrieval in the
     * respective <code>logCalcChanges</code>.
     * 
     * This works under the assumption that there is <em>no</em> overlap
     * between UserActions on different <code>Worksheets</code>.
     * 
     * @param dialog
     *            "worksheet" | "exp" | "thy" | "pbl" | "met"
     * @param worksheet_id
     *            is "" for "exp" & "thy" & "pbl" & "met"
     * @param action
     *            IUserAction | "LO_START_SESSION" ... | "LO_OPEN_WORKSHEET"
     */
    public void logUserAction(String dialog, IUserAction action) {
        switch (action.getActionID()) {
        case UI_MINIBROWSER_LINK: {
            // for the previous step there already was a stepToDatabase
            // (incl.current_step_.reset)
            EUIElement action_name;
            String argument;
            if (isStartExp((UserActionOnLink) action)) {
                action_name = isac.useractions.EUIElement.LO_OPEN_WORKSHEET;
            } else {
                action_name = action.getActionID();
            }
            current_step_.setAction(user_.getUsername(), session_id_, dialog,
                    action_name, kestoreKeyOf(((UserActionOnLink) action)
                            .getLink().toString()), null/* subst */);
            // TOOODO if context then current_step_ completed in
            // UserLogger#logContext???
            // else currentStepToDatabase() !
            return;
        }
        case UI_SOLVE_CALCULATE_ALL:
        case UI_SOLVE_CALCULATE_1: {
            current_step_.setAction(user_.getUsername(), session_id_, dialog,
                    action.getActionID(), null, null);
            // TOOODO if context then current_step_ completed in
            // UserLogger#logContext???
            // else currentStepToDatabase() !
            return;
        }
        case UI_SOLVE_EDIT_ACTIVE_FORMULA_COMPLETE: {
            String input = ((UserActionOnCalcElement) action).getCalcElement()
                    .toSMLString();
            current_step_.setAction(user_.getUsername(), session_id_, dialog,
                    action.getActionID(), input, null);
        }
        // TODO javacode?? "otherwise" // don't record anything
        // in this case also no calcChanged occurs
        }
    }

    /**
     * log closing the worksheet; if the respective calculation has never been
     * finished before, first log LO_STOP_EXAMPLE with failure of that example.
     */
    public void logCloseWorksheet(String worksheet_id,
            boolean calculation_finished_successfully) {
        if (!calculation_finished_successfully) {
            current_step_.setAction(user_.getUsername(), session_id_,
                    "worksheet", isac.useractions.EUIElement.LO_STOP_EXAMPLE,
                    null, null);
            current_step_.setResult(worksheet_id, null, null, null, new Float(
                    0.0));
            currentStepToDatabase();
            current_step_.reset();
            this.exp_dialog_.notifySomeExampleFinished();
        }
        current_step_.setAction(user_.getUsername(), session_id_, "worksheet",
                isac.useractions.EUIElement.UI_CLOSE_WORKSHEET, null, null);
        current_step_.setResult(worksheet_id, null, null, null, null);
        currentStepToDatabase();
    }

    /**
     * WN0710 the similar UI_CLOSE_WORKSHEET is done in the gui; analogously
     * this could better be done on closing WindowApplication ?
     */
    public void logStopSession() {
        current_step_.setAction(user_.getUsername(), session_id_, null,
                isac.useractions.EUIElement.UI_STOP_SESSION, null, null);
        currentStepToDatabase();
    }

    /**
     * Complete a <code>UserRecord</code> initialized in
     * <code>logUserAction</code> with the related data returned from the
     * MathEngine in the <code>CalcChanged</code> which always indicates
     * success.
     */
    public void logCalcChanged(String worksheet_id, CalcChanged event) {
        // worksheet_map_success_.put(worksheet_id, new SuccessInfo((float)
        // 1.0));
        String position = "RemoteException";
        String formula_from = "RemoteException";
        String formula_to = "RemoteException";
        boolean end_of_calculation = false;
        try {
            position = event.getLastUnchangedFormula().toSMLString();

            if (event.getLastUnchangedFormula().onCalcHead())
                formula_from = ((Formula) ((CalcHead) event
                        .getLastUnchangedFormula().getElement()).getHeadLine())
                        .toSMLString();
            else
                formula_from = (((Formula) event.getLastUnchangedFormula()
                        .getElement()).toSMLString());

            if (event.getLastGeneratedFormula().onCalcHead())
                formula_to = ((Formula) ((CalcHead) event
                        .getLastGeneratedFormula().getElement()).getHeadLine())
                        .toSMLString();
            else
                formula_to = (((Formula) event.getLastGeneratedFormula()
                        .getElement()).toSMLString());

            if (event.getLastGeneratedFormula().getPosition()
                    .onEndOfCalculation()) // remote call
                end_of_calculation = true;

        } catch (RemoteException e1) {
            e1.printStackTrace();
        }
        current_step_.setResult(worksheet_id, position, formula_from,
                formula_to, new Float(1.0));// TODO use SuccessInfo not 1.0
        currentStepToDatabase();
        if (end_of_calculation) {
            current_step_.setAction(user_.getUsername(), session_id_,
                    "worksheet", isac.useractions.EUIElement.LO_STOP_EXAMPLE,
                    null, null);
            current_step_.setResult(worksheet_id, null, null, null, new Float(
                    1.0));// TODO use SuccessInfo not 1.0
            currentStepToDatabase();
            this.exp_dialog_.notifySomeExampleFinished();
        }
    }

    /**
     * Complete a <code>UserRecord</code> initialized in
     * <code>logUserAction</code> with the related data returned from the
     * MathEngine in the <code>CalcMessage</code> which always indicates
     * failure.
     */
    public void logCalcMessage(String worksheet_id, String position,
            String message) {
        current_step_.setResult(worksheet_id, position, null, message,
                new Float(0.0));// TODO use SuccessInfo not 1.0
        currentStepToDatabase();
    }

    /**
     * Return the <code>KEStoreKey</code> 'thy_xxx.xml' of an url-string like
     * www/kbase/exp/thy_xxx.html#COMMAND_EXEC_EXAMPLE#thy_xxx.xml or
     * <code>null</code> otherwise
     * 
     * WN 070908 TOOODO: make this a method of some new? class ?IsacLink?
     * 
     * @param <paramname>
     *            TOOODO <paramdescription>
     * @return <code>KEStoreKey</code>
     */
    private String kestoreKeyOf(String url) {
        String[] link_array; // ..,proto2/www/kbase/exp/thy_xxx.html#COMMAND_EXEC_EXAMPLE#thy_xxx.xml
        String command = ""; // COMMAND_EXEC_EXAMPLE

        link_array = url.split("#");
        if (link_array.length >= 2) {
            command = link_array[1];
            if (command.equals("COMMAND_EXEC_EXAMPLE"))
                return link_array[2];
        }
        return null;
    }

    /**
     * Is the <code>UserActionOnLink</code> meant to start an example ? In
     * this case the link looks like url.html#COMMAND_EXEC_EXAMPLE#example.xml
     * 
     * WN 070908 TOOODO: make this a method of some new? class ?IsacLink?
     */
    private boolean isStartExp(UserActionOnLink action) {
        String[] link_array = action.getLink().toString().split("#");
        return link_array.length >= 3;
    }

    /**
     * @deprecated quick and dirty made Compod run.
     */
    public void setExampleDialog(ExampleDialog exp_dialog) {
        exp_dialog_ = exp_dialog;
    }
}
